##############################################################################
# General library functions and variable definitions
##############################################################################

require 'open3'

module TwilioHelper

  BASE_DIR = ENV['BASE_DIR']
  BASE_VERSION = ENV['BASE_VERSION']
  SDK_RELEASE_VERSION = ENV['SDK_RELEASE_VERSION']
  PRERELEASE_QUALIFIER = ENV['PRERELEASE_QUALIFIER']
  SDK_SOURCE_CODE_REVISION = ENV['SDK_SOURCE_CODE_REVISION']

  VERSION_FILE = "#{BASE_DIR}/version.txt"
  RC_VERSION_FILE = "#{BASE_DIR}/version-rc.txt"

  CHANGELOG_FILE = "#{BASE_DIR}/CHANGELOG.md"
  CHANGELOG_SEPARATOR = '----------'
  CHANGELOG_KNOWN_ISSUES_HEADER = 'Known Issues'

  def self.fail_and_exit code, message
    puts message
    exit code
  end

  def self.execute_shell_command command
    command_status = false

    Open3.popen2e(command) do |stdin, stdout_err, wait_thr|
      while line = stdout_err.gets
        puts line
      end

      command_status = wait_thr.value
    end

    return command_status == 0 ? true : false
  end

  def self.version_file
    VERSION_FILE
  end

  def self.rc_version_file
    RC_VERSION_FILE
  end

  def self.changelog_file
    CHANGELOG_FILE
  end

  def self.scm_commit
    SDK_SOURCE_CODE_REVISION
  end

  # Snapshot version helpers
  def self.snapshot_version
    "#{BASE_VERSION}-#{SDK_SOURCE_CODE_REVISION}"
  end

  # RC version helpers
  def self.rc_number
    File.open(RC_VERSION_FILE, &:readline).strip
  end

  def self.rc_version
    "#{BASE_VERSION}-rc#{rc_number}"
  end

  def self.increment_rc_number
    new_rc_number = rc_number.to_i + 1
    File.open(rc_version_file, "w") {|file| file.puts new_rc_number}
  end

  def self.reset_rc_number
    File.open(rc_version_file, "w") {|file| file.puts '1'}
  end

  # Release version helpers
  def self.release_version
    SDK_RELEASE_VERSION
  end

  def self.is_prerelease? 
    is_prerelease = false
    prerelease_qualifier = PRERELEASE_QUALIFIER unless PRERELEASE_QUALIFIER.nil? || PRERELEASE_QUALIFIER.length == 0

    unless prerelease_qualifier.nil?
      is_prerelease = !prerelease_qualifier.match(/(preview|beta)([0-9]+)/).nil?
    end

    is_prerelease
  end

  def self.core_version
    `mvn -q -Dbuild.platform=ios -Dexec.executable='echo' -Dexec.args='${dependencies.twilio-video.version}' --non-recursive exec:exec`.strip
  end

  def self.release_description version
    release_description = "This release consumes twilio-video-cpp-#{core_version}.\n\n"
    release_description += changelog_contents(version).gsub(/#### #{version} \(In Development\)\n\n/, '')
    release_description
  end

  def self.next_version
    base_version = BASE_VERSION
    prerelease_qualifier = PRERELEASE_QUALIFIER unless PRERELEASE_QUALIFIER.nil? || PRERELEASE_QUALIFIER.length == 0

    # If we have a prerelease qualifier, update that
    unless prerelease_qualifier.nil?
      matches = prerelease_qualifier.match(/(preview|beta)([0-9]+)/)

      unless matches.nil?
        prefix, number = matches.captures
        prerelease_qualifier = "#{prefix}#{number.to_i + 1}"
      end
    else
      matches = base_version.match(/([0-9]+).([0-9]+).([0-9]+)/)

      unless matches.nil?
        major, minor, patch = matches.captures
        base_version = "#{major}.#{minor}.#{patch.to_i + 1}"
      end
    end

    return base_version, prerelease_qualifier
  end

  def self.increment_version
    base_version, prerelease_qualifier = next_version

    File.open(version_file, "w")  do |file|
      file.puts base_version
      file.puts prerelease_qualifier unless prerelease_qualifier.nil?
    end
  end

  # Changelog helpers
  def self.changelog_contents version
    changelog_contents = ""
    version_found = false

    File.open(changelog_file).each do |line|
      if line.match(/#{version}/i)
        version_found = true
      end

      if version_found
        if line.match(/#{CHANGELOG_SEPARATOR}/)
          break
        end

        changelog_contents += line
      end
    end

    version_found ? changelog_contents : nil
  end

  def self.known_issues version
    known_issues = ""
    version_found = false
    known_issues_found = false

    File.open(changelog_file).each do |line|
      if line.match(/#{version}/i)
        version_found = true
      end

      if version_found
        if line.match(/#{CHANGELOG_KNOWN_ISSUES_HEADER}/)
          known_issues_found = true
        end

        if known_issues_found
          if line.match(/#{CHANGELOG_SEPARATOR}/)
            break
          end

          known_issues += line
        end
      end
    end

    known_issues_found ? known_issues : nil
  end

  def self.update_changelog
    current_known_issues = known_issues release_version
    next_base_version, next_prerelease_qualifier = next_version
    next_version = next_base_version + (next_prerelease_qualifier ? "-#{next_prerelease_qualifier}" : '')

    current_changelog = File.readlines(changelog_file)
    new_changelog_entry = "#### #{next_version} (In Development)\n\n#{current_known_issues}#{CHANGELOG_SEPARATOR}\n\n"

    File.open(changelog_file, 'r+') do |file|
      current_changelog.each do |line|
        if line.match(/#{release_version} \(In Development\)/i)
          file.puts new_changelog_entry
          file.puts line.gsub(/In Development/, Time.now.strftime('%B %-e, %Y'))
          next
        end

        file.puts line
      end
    end
  end

end # module TwilioHelper